#include <stdint.h>
#include <stdio.h>

#define DEFAULT_TAPS 0xb8
#define WALL_BIT 7
#define WALL_ORIENT 0
#define UNDERGROUND_MOVE 3

// Period 255 taps: 0x8e, 0xb8 [Pitfall], 0x95, 0xd4, 0x96, 0xb4, 0xa6,
// 0xb2, 0xaf, 0xfa, 0xb1, 0xc6, 0xc3, 0xe1, 0xe7, 0xf3
// Direction: 0: Walk right
// 	      1 or 2: Walk left
uint8_t lfsr(uint8_t random, int taps, int direction) {
	uint8_t t, r, q; 
	q = taps;
	r = random;
	t = 0;
	if(direction == 0) {
		taps &= 0x7f;
		taps <<= 1;
		taps |= 1;
	}
	while(taps > 0) {
		if(taps & 1) {
			t ^= r;
		}
		r >>= 1;
		taps >>= 1;
	}
	if(direction < 1) {
		random >>= 1;
		random |= t << 7;
	} else {
		random <<= 1;
		random |= (t & 1);
	}
	return random;
}

void describe_pitfall_room(uint8_t random) {
	uint8_t scene_type;
	scene_type = (random >> 3) & 7;
	switch(scene_type) {
		case 1: printf("Two holes around ");
		case 0: printf("Hole with ladder, ");
			break;
		case 2: printf("Tar pit with vine");
			break;
		case 3: printf("Quicksand with vine");
			break;
		case 4: printf("Crocodile pit");
			break;
		case 5: printf("Shifting tar pit and TREASURE: ");
			break;
		case 6: printf("Shifting tar pit with vine");
			break;
		case 7: printf("Shifting quicksand");
	}
	if(scene_type == 5) {
		switch(random & 3) {
			case 0: printf("Money bag");
				break;
			case 1: printf("Silver brick");
				break;
			case 2: printf("Gold brick");
				break;
			case 3: printf("Diamond ring");
		}
	} else if((scene_type & 6) == 0) {
		switch((random >> WALL_BIT) & 1) {
			case WALL_ORIENT: printf("wall on left");
				break;
			default: printf("wall on right");
		}
	}
	if(scene_type != 4 && scene_type != 5) {
		printf(" with ");
		switch(random & 7) {
			case 0: printf("1 rolling log"); break;
			case 1: printf("2 rolling logs near"); break;
			case 2: printf("2 rolling logs far"); break;
			case 3: printf("3 rolling logs"); break;
			case 4: printf("1 still log"); break;
			case 5: printf("3 still logs"); break;
			case 6: printf("Fire"); break;
			case 7: printf("Rattlesnake"); 
		}
	}
	printf("\n");
}

// Given a point and a count, count the number of treasures we walk past
// when going underground
int count_treasures(int random, int taps) {
	uint8_t scene_type;
	int out = 0, loopstop;	
	scene_type = (random >> 3) & 7;
	if(((scene_type&6) != 0) || (((random>>WALL_BIT)&1)!=WALL_ORIENT)) {
		return -1; // Must start with wall on left
	}
	for(loopstop = 0;loopstop < 500;loopstop++) {
		random = lfsr(random, taps, 2); // Move right
		scene_type = (random >> 3) & 7;
		if(scene_type == 5) {
			out++;
		}
		if(((loopstop % 3) == 2) && (scene_type & 6) == 0) {
			if((((random>>WALL_BIT)&1)==WALL_ORIENT)) {
				return -2; // Dead end
			}
			return out;
		}
	} 
	return -3; // Unexpected error
}

// Given a place in the jungle, find where the exit is for this place
int find_exit(int random, int taps, int place, int dir1) {
	int a, dir;
	int max = 0;
	int cost = 0; // Number of rooms to travel underground to get to exit
	if((random & 0x30) != 0) {
		printf("XX ");
		return -1; // Not an enterance, no exit
	}
	switch((random >> WALL_BIT) & 1) {
		case WALL_ORIENT: dir = UNDERGROUND_MOVE; break; // Wall onleft
		default: dir = -UNDERGROUND_MOVE; // Wall on right
	}	
	do {
		if(dir > 0) {
			for(a = 0; a < UNDERGROUND_MOVE; a++) {
				random = lfsr(random, taps, 2);
			}
		} else {
			for(a = 0; a < UNDERGROUND_MOVE; a++) {
				random = lfsr(random, taps, 0);
			}
		}
		place += 255; // Avoid % operator on negative numbers
		if(dir1 == 0) {
			place += dir;
		} else {
			place -= dir;
		}
		place %= 255;
		cost++;
		if((random & 0x30) == 0) { // If this is an exit
			a = (random >> WALL_BIT) & 1;
			if(dir == UNDERGROUND_MOVE && a == WALL_ORIENT) {
				printf("|| "); 
				return -2; // Dead end
			} else if(dir == -UNDERGROUND_MOVE && a !=WALL_ORIENT){
				printf("|| "); 
				return -2; // Dead end
			}
			printf("%02x,%x ", place + 1, cost);
			return place;	
		}
		max++;
	} while(random != 0 && max < 512);
	printf("XX ");
	return -3; // Invalid value
}
	
int main(int argc, char **argv) {
	int max, taps, direction;
	uint8_t random = 0xc4;
	uint8_t lrandom = 0xc4;
	int dir1 = 0; int dir2 = 2, q, npassage = 0, ntreas = 0;
	if(argc == 1) {
		taps = DEFAULT_TAPS;
	} else if(argc == 2 || argc == 3) {
		taps = strtol(argv[1],NULL,16);
		if(argc == 3) { // Render map as RtoLeft, not Left2Right
			dir1 = 2;
			dir2 = 0;
		}
	} else {
		printf("Usage: lfsr_pitfall [taps], where taps can have the "
"value\n0x8e, 0xb8 [Pitfall], 0x95, 0xd4, 0x96, 0xb4, 0xa6, 0xb2, 0xaf, "
"0xfa,\n0xb1, 0xc6, 0xc3, 0xe1, 0xe7, 0xf3\n");
		return 1;
	}
	if(taps >= 512) {
		for(direction=0;direction<=2;direction++) {
		    for(taps=0;taps<=255;taps++) {
			printf("%d %02x - ",direction,taps);
			lrandom = 0xc4;
			max = 0;
			do {
				printf("%02x ",lrandom);
				lrandom = lfsr(lrandom, taps, direction);
				max++;
			} while(lrandom != 0xc4 && lrandom != 0 && max < 256);
			printf("\n");
		    }
		}
	} else if(taps >= 128) {
		lrandom = 0xc4;
		random = 0xc4;
		max = 0;
		do {
			printf("%02x ",max + 1);
			if((random & 0x30) == 0) {
				q = find_exit(random, taps, max, dir1);
				if(q > max) {
					npassage++;
				}
				if(count_treasures(random, taps)
						 == 0) {
					ntreas++;
				}
			} else {
				if((random & 0x38) == 0x28) { // Treasure
					printf("** ");
				} else {
					printf("-- ");
				}
			}
			describe_pitfall_room(random);
			lrandom = lfsr(lrandom, taps, dir1);
			random = lfsr(random, taps, dir2);
			max++;
		} while(lrandom != 0xc4 && lrandom != 0 && max < 256);
		printf("Number of passages %d\n",npassage);
		printf("# of passages that do not skip treasure %d\n",ntreas);
	}
}

